@@ -27,8 +27,9 @@
 #define _COMPIZ_CORE_H
 
 #include <compiz-plugin.h>
+#include <dlfcn.h>
 
-#define CORE_ABIVERSION 20090619
+#define CORE_ABIVERSION 20180616
 
 #include <stdio.h>
 #include <sys/time.h>
@@ -217,8 +218,9 @@ extern Bool       indirectRendering;
 extern Bool       strictBinding;
 extern Bool       useCow;
 extern Bool       noDetection;
-extern Bool	  useDesktopHints;
+extern Bool       useDesktopHints;
 extern Bool       onlyCurrentScreen;
+extern Bool       noFBO;
 
 extern char	**initialPlugins;
 extern int 	nInitialPlugins;
@@ -773,9 +775,11 @@ removeFileWatch (CompFileWatchHandle handle);
 #define COMP_DISPLAY_OPTION_HIDE_SKIP_TASKBAR_WINDOWS        29
 #define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_SHADED_KEY         30
 #define COMP_DISPLAY_OPTION_IGNORE_HINTS_WHEN_MAXIMIZED      31
-#define COMP_DISPLAY_OPTION_PING_DELAY			     32
+#define COMP_DISPLAY_OPTION_PING_DELAY                       32
 #define COMP_DISPLAY_OPTION_EDGE_DELAY                       33
-#define COMP_DISPLAY_OPTION_NUM				     34
+#define COMP_DISPLAY_OPTION_CURSOR_THEME                     34
+#define COMP_DISPLAY_OPTION_CURSOR_SIZE                      35
+#define COMP_DISPLAY_OPTION_NUM                              36
 
 typedef void (*HandleEventProc) (CompDisplay *display,
 				 XEvent	     *event);
@@ -975,6 +979,7 @@ struct _CompDisplay {
     Atom clientListStackingAtom;
 
     Atom frameExtentsAtom;
+    Atom gtkFrameExtentsAtom;
     Atom frameWindowAtom;
 
     Atom wmStateAtom;
@@ -1028,6 +1033,7 @@ struct _CompDisplay {
     GLenum textureFilter;
 
     Window activeWindow;
+    Window nextActiveWindow;
 
     Window below;
     char   displayString[256];
@@ -1121,6 +1127,9 @@ logMessage (const char   *componentName,
 const char *
 logLevelToString (CompLogLevel level);
 
+void
+launchFallbackWM (void);
+
 int
 compCheckForError (Display *dpy);
 
@@ -1677,15 +1686,14 @@ disableTexture (CompScreen  *screen,
 #define COMP_SCREEN_OPTION_UNREDIRECT_FS	  5
 #define COMP_SCREEN_OPTION_DEFAULT_ICON		  6
 #define COMP_SCREEN_OPTION_SYNC_TO_VBLANK	  7
-#define COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS	  8
-#define COMP_SCREEN_OPTION_DETECT_OUTPUTS	  9
-#define COMP_SCREEN_OPTION_OUTPUTS		  10
-#define COMP_SCREEN_OPTION_OVERLAPPING_OUTPUTS	  11
-#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_LEVEL 12
-#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_MATCH 13
-#define COMP_SCREEN_OPTION_TEXTURE_COMPRESSION	  14
-#define COMP_SCREEN_OPTION_FORCE_INDEPENDENT      15
-#define COMP_SCREEN_OPTION_NUM		          16
+#define COMP_SCREEN_OPTION_DETECT_OUTPUTS	  8
+#define COMP_SCREEN_OPTION_OUTPUTS		  9
+#define COMP_SCREEN_OPTION_OVERLAPPING_OUTPUTS	  10
+#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_LEVEL 11
+#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_MATCH 12
+#define COMP_SCREEN_OPTION_TEXTURE_COMPRESSION	  13
+#define COMP_SCREEN_OPTION_FORCE_INDEPENDENT      14
+#define COMP_SCREEN_OPTION_NUM		          15
 
 #ifndef GLX_EXT_texture_from_pixmap
 #define GLX_BIND_TO_TEXTURE_RGB_EXT        0x20D0
@@ -2590,6 +2598,7 @@ struct _CompWindow {
     Bool	      damaged;
     Bool	      redirected;
     Bool	      managed;
+    Bool	      unmanaging;
     Bool	      bindFailed;
     Bool	      overlayWindow;
     int		      destroyRefCnt;
@@ -2644,11 +2653,14 @@ struct _CompWindow {
     XWindowChanges saveWc;
     int		   saveMask;
 
+    /* For saving changes before fullscreen */
+    XWindowChanges saveFFWc;
+    int		   savedFF;
+
     XSyncCounter  syncCounter;
     XSyncValue	  syncValue;
     XSyncAlarm	  syncAlarm;
     unsigned long syncAlarmConnection;
-    unsigned int  syncWaitHandle;
 
     Bool syncWait;
     int	 syncX;
@@ -2677,7 +2689,9 @@ struct _CompWindow {
     /* must be set by addWindowGeometry */
     DrawWindowGeometryProc drawWindowGeometry;
 
-    void *reserved;
+    /* placed here to preserve ABI compat */
+    CompWindowExtents clientFrame;
+    CompWindowExtents frameInput;
 };
 
 #define GET_CORE_WINDOW(object) ((CompWindow *) (object))
@@ -2805,6 +2819,9 @@ updateTransientHint (CompWindow *w);
 void
 updateIconGeometry (CompWindow *w);
 
+void
+updateClientFrame (CompWindow *w);
+
 Window
 getClientLeader (CompWindow *w);
 
@@ -2828,6 +2845,10 @@ void
 setWindowFrameExtents (CompWindow	 *w,
 		       CompWindowExtents *input);
 
+void
+setWindowBorderExtents (CompWindow	  *w,
+			CompWindowExtents *input);
+
 void
 updateWindowOutputExtents (CompWindow *w);
 
@@ -3385,9 +3406,9 @@ matchPropertyChanged (CompDisplay *display,
 #define RESTOSTRING(min, max) MINTOSTRING (min) MAXTOSTRING (max)
 
 typedef struct _CompMetadataOptionInfo {
-    char		   *name;
-    char		   *type;
-    char		   *data;
+    const char		   *name;
+    const char		   *type;
+    const char		   *data;
     CompActionCallBackProc initiate;
     CompActionCallBackProc terminate;
 } CompMetadataOptionInfo;
