@@ -48,10 +48,10 @@ typedef struct _Vector {
     int	y0;
 } Vector;
 
+/* FIXME: Remove */
 #define DECOR_BARE   0
-#define DECOR_NORMAL 1
-#define DECOR_ACTIVE 2
-#define DECOR_NUM    3
+#define DECOR_ACTIVE 1
+#define DECOR_NUM    2
 
 typedef struct _DecorTexture {
     struct _DecorTexture *next;
@@ -65,10 +65,15 @@ typedef struct _Decoration {
     int		      refCount;
     DecorTexture      *texture;
     CompWindowExtents output;
-    CompWindowExtents input;
-    CompWindowExtents maxInput;
+    CompWindowExtents frame;
+    CompWindowExtents border;
+    CompWindowExtents maxFrame;
+    CompWindowExtents maxBorder;
     int		      minWidth;
     int		      minHeight;
+    unsigned int      frameType;
+    unsigned int      frameState;
+    unsigned int      frameActions;
     decor_quad_t      *quad;
     int		      nQuad;
 } Decoration;
@@ -113,7 +118,9 @@ typedef struct _DecorDisplay {
     DecorTexture	     *textures;
     Atom		     supportingDmCheckAtom;
     Atom		     winDecorAtom;
+    Atom		     requestFrameExtentsAtom;
     Atom		     decorAtom[DECOR_NUM];
+    Atom		     inputFrameAtom;
 
     CompOption opt[DECOR_DISPLAY_OPTION_NUM];
 } DecorDisplay;
@@ -123,11 +130,13 @@ typedef struct _DecorScreen {
 
     Window dmWin;
 
-    Decoration *decor[DECOR_NUM];
+    Decoration   **decors[DECOR_NUM];
+    unsigned int decorNum[DECOR_NUM];
 
     DrawWindowProc		  drawWindow;
     DamageWindowRectProc	  damageWindowRect;
     GetOutputExtentsForWindowProc getOutputExtentsForWindow;
+    AddSupportedAtomsProc         addSupportedAtoms;
 
     WindowMoveNotifyProc   windowMoveNotify;
     WindowResizeNotifyProc windowResizeNotify;
@@ -139,7 +148,10 @@ typedef struct _DecorScreen {
 
 typedef struct _DecorWindow {
     WindowDecoration *wd;
-    Decoration	     *decor;
+    Decoration	     **decors;
+    unsigned int     decorNum;
+
+    Window	     inputFrame;
 
     CompTimeoutHandle resizeUpdateHandle;
 } DecorWindow;
@@ -188,6 +200,11 @@ decorDrawWindow (CompWindow	      *w,
     status = (*w->screen->drawWindow) (w, transform, attrib, region, mask);
     WRAP (ds, w->screen, drawWindow, decorDrawWindow);
 
+    /* we wait to draw dock shadows until we get to the lowest
+       desktop window in the stack */
+    if (w->type & CompWindowTypeDockMask)
+	return status;
+
     if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
 	region = &infiniteRegion;
 
@@ -224,6 +241,65 @@ decorDrawWindow (CompWindow	      *w,
 					     attrib, mask);
     }
 
+    if (w->type & CompWindowTypeDesktopMask)
+    {
+	/* we only want to draw on the lowest desktop window, find it and see
+	   if we the window we have is it */
+	CompWindow *window = w->screen->windows;
+	for (window = w->screen->windows; window; window = window->next)
+	{
+	    if (window->type & CompWindowTypeDesktopMask)
+	    {
+		if (window == w)
+		    break;
+		else
+		    return status;
+	    }
+	}
+
+	/* drawing dock shadows now */
+	for (window = w->screen->windows; window; window = window->next)
+	{
+	    if (window->type & CompWindowTypeDockMask && !window->destroyed && !window->invisible)
+	    {
+		DECOR_WINDOW (window);
+
+		if (dw->wd && region->numRects)
+		{
+		    WindowDecoration *wd = dw->wd;
+		    REGION	     box;
+		    int		     i;
+
+		    mask |= PAINT_WINDOW_BLEND_MASK;
+
+		    box.rects	 = &box.extents;
+		    box.numRects = 1;
+
+		    window->vCount = window->indexCount = 0;
+
+		    for (i = 0; i < wd->nQuad; i++)
+		    {
+			box.extents = wd->quad[i].box;
+
+			if (box.extents.x1 < box.extents.x2 &&
+			    box.extents.y1 < box.extents.y2)
+			{
+			    (*window->screen->addWindowGeometry) (window,
+								  &wd->quad[i].matrix, 1,
+								  &box,
+								  region);
+			}
+		    }
+
+		    if (window->vCount)
+			(*window->screen->drawWindowTexture) (window,
+							      &wd->decor->texture->texture,
+							      attrib, mask);
+		}
+	    }
+	}
+    }
+
     return status;
 }
 
@@ -388,19 +464,18 @@ computeQuadBox (decor_quad_t *q,
 }
 
 static Decoration *
-decorCreateDecoration (CompScreen *screen,
-		       Window	  id,
-		       Atom	  decorAtom)
+decorCreateDecoration (CompScreen   *screen,
+		       Window	    id,
+		       long	    *prop,
+		       unsigned int size,
+		       unsigned int type,
+		       unsigned int nOffset)
 {
     Decoration	    *decoration;
-    Atom	    actual;
-    int		    result, format;
-    unsigned long   n, nleft;
-    unsigned char   *data;
-    long	    *prop;
+    unsigned int    frameType, frameState, frameActions;
     Pixmap	    pixmap;
-    decor_extents_t input;
-    decor_extents_t maxInput;
+    decor_extents_t frame, border;
+    decor_extents_t maxFrame, maxBorder;
     decor_quad_t    *quad;
     int		    nQuad;
     int		    minWidth;
@@ -408,52 +483,31 @@ decorCreateDecoration (CompScreen *screen,
     int		    left, right, top, bottom;
     int		    x1, y1, x2, y2;
 
-    result = XGetWindowProperty (screen->display->display, id,
-				 decorAtom, 0L, 1024L, FALSE,
-				 XA_INTEGER, &actual, &format,
-				 &n, &nleft, &data);
+    nQuad = N_QUADS_MAX;
 
-    if (result != Success || !data)
-	return NULL;
-
-    if (!n)
-    {
-	XFree (data);
-	return NULL;
-    }
-
-    prop = (long *) data;
-
-    if (decor_property_get_version (prop) != decor_version ())
-    {
-	compLogMessage ("decoration", CompLogLevelWarn,
-			"Property ignored because "
-			"version is %d and decoration plugin version is %d\n",
-			decor_property_get_version (prop), decor_version ());
-
-	XFree (data);
-	return NULL;
-    }
-
-    nQuad = (n - BASE_PROP_SIZE) / QUAD_PROP_SIZE;
-
-    quad = malloc (sizeof (decor_quad_t) * nQuad);
+    quad = calloc (nQuad, sizeof (decor_quad_t));
     if (!quad)
     {
-	XFree (data);
+	compLogMessage ("decoration",
+			CompLogLevelWarn,
+			"Failed to allocate %i quads\n", nQuad);
 	return NULL;
     }
 
-    nQuad = decor_property_to_quads (prop,
-				     n,
-				     &pixmap,
-				     &input,
-				     &maxInput,
-				     &minWidth,
-				     &minHeight,
-				     quad);
-
-    XFree (data);
+    nQuad = decor_pixmap_property_to_quads (prop,
+					    nOffset,
+					    size,
+					    &pixmap,
+					    &frame,
+					    &border,
+					    &maxFrame,
+					    &maxBorder,
+					    &minWidth,
+					    &minHeight,
+					    &frameType,
+					    &frameState,
+					    &frameActions,
+					    quad);
 
     if (!nQuad)
     {
@@ -508,15 +562,34 @@ decorCreateDecoration (CompScreen *screen,
     decoration->output.top    = -top;
     decoration->output.bottom = bottom - minHeight;
 
-    decoration->input.left   = input.left;
-    decoration->input.right  = input.right;
-    decoration->input.top    = input.top;
-    decoration->input.bottom = input.bottom;
+    /* Extents of actual frame window */
+    decoration->frame.left   = frame.left;
+    decoration->frame.right  = frame.right;
+    decoration->frame.top    = frame.top;
+    decoration->frame.bottom = frame.bottom;
 
-    decoration->maxInput.left   = maxInput.left;
-    decoration->maxInput.right  = maxInput.right;
-    decoration->maxInput.top    = maxInput.top;
-    decoration->maxInput.bottom = maxInput.bottom;
+    /* Border extents */
+    decoration->border.left   = border.left;
+    decoration->border.right  = border.right;
+    decoration->border.top    = border.top;
+    decoration->border.bottom = border.bottom;
+
+    /* Extents of actual frame window */
+    decoration->maxFrame.left   = maxFrame.left;
+    decoration->maxFrame.right  = maxFrame.right;
+    decoration->maxFrame.top    = maxFrame.top;
+    decoration->maxFrame.bottom = maxFrame.bottom;
+
+    /* Border extents */
+    decoration->maxBorder.left   = maxBorder.left;
+    decoration->maxBorder.right  = maxBorder.right;
+    decoration->maxBorder.top    = maxBorder.top;
+    decoration->maxBorder.bottom = maxBorder.bottom;
+
+    /* Decoration info */
+    decoration->frameType    = frameType;
+    decoration->frameState   = frameState;
+    decoration->frameActions = frameActions;
 
     decoration->refCount = 1;
 
@@ -527,8 +600,11 @@ static void
 decorReleaseDecoration (CompScreen *screen,
 			Decoration *decoration)
 {
+    if (!decoration)
+	return;
+
     decoration->refCount--;
-    if (decoration->refCount)
+    if (decoration->refCount > 0)
 	return;
 
     decorReleaseTexture (screen, decoration->texture);
@@ -537,20 +613,150 @@ decorReleaseDecoration (CompScreen *screen,
     free (decoration);
 }
 
+static void
+decorReleaseDecorations (CompScreen   *screen,
+			 Decoration   **decors,
+			 unsigned int *decorNum)
+{
+    int i;
+
+    if (decorNum && *decorNum > 0)
+    {
+	if (decors)
+	{
+	    for (i = 0; i < *decorNum; i++)
+	    {
+		if (decors[i])
+		    decorReleaseDecoration (screen, decors[i]);
+	    }
+	    free (decors);
+	}
+	*decorNum = 0;
+    }
+}
+
+static Decoration **
+decorUpdateDecorations (CompScreen   *screen,
+			Window       id,
+			Atom	     decorAtom,
+			Decoration   **decors,
+			unsigned int *decorNum)
+{
+    int		    i;
+    unsigned long   n, nleft;
+    unsigned char   *data;
+    long	    *prop;
+    Atom	    actual;
+    int		    result, format;
+    unsigned int    type;
+
+    /* null pointer as decorNum is unacceptable */
+    if (!decorNum)
+	return NULL;
+
+    /* FIXME: should really check the property to see
+       if the decoration changed, and /then/ release
+       and re-update it, but for now just releasing all */
+    if (decors && *decorNum > 0)
+	decorReleaseDecorations (screen, decors, decorNum);
+    decors = NULL;
+    *decorNum = 0;
+
+    result = XGetWindowProperty (screen->display->display, id,
+				 decorAtom, 0L,
+				 PROP_HEADER_SIZE + 6 * (BASE_PROP_SIZE +
+				   QUAD_PROP_SIZE * N_QUADS_MAX),
+				 FALSE, XA_INTEGER, &actual, &format,
+				 &n, &nleft, &data);
+
+    if (result != Success || !data)
+	return NULL;
+
+    if (!n)
+    {
+	XFree (data);
+	return NULL;
+    }
+
+    /* attempted to read the reasonable amount of
+     * around 6 pixmap decorations, if there are more, we need
+     * an additional roundtrip to read everything else
+     */
+    if (nleft)
+    {
+	XFree (data);
+
+	result = XGetWindowProperty (screen->display->display, id, decorAtom, 0L,
+				     n + nleft, FALSE, XA_INTEGER,
+				     &actual, &format,
+				     &n, &nleft, &data);
+
+	if (result != Success || !data)
+	    return NULL;
+
+	if (!n)
+	{
+	    XFree (data);
+	    return NULL;
+	}
+    }
+
+    prop = (long *) data;
+
+    if (decor_property_get_version (prop) != decor_version ())
+    {
+	compLogMessage ("decoration", CompLogLevelWarn,
+			"Property ignored because "
+			"version is %d and decoration plugin version is %d\n",
+			decor_property_get_version (prop), decor_version ());
+
+	XFree (data);
+	return NULL;
+    }
+
+    type = decor_property_get_type (prop);
+
+    *decorNum = decor_property_get_num (prop);
+    if (*decorNum == 0)
+    {
+	XFree (data);
+	return NULL;
+    }
+
+    /* create a new decoration for each individual item on the property */
+    decors = calloc (*decorNum, sizeof (Decoration *));
+    for (i = 0; i < *decorNum; i++)
+	decors[i] = NULL;
+    for (i = 0; i < *decorNum; i++)
+    {
+	Decoration *d = decorCreateDecoration (screen, id, prop, n, type, i);
+
+	if (d)
+	    decors[i] = d;
+	else
+	{
+	    unsigned int realDecorNum = i;
+	    decorReleaseDecorations (screen, decors, &realDecorNum);
+	    *decorNum = realDecorNum;
+	    XFree (data);
+	    return NULL;
+	}
+    }
+
+    XFree (data);
+
+    return decors;
+}
+
 static void
 decorWindowUpdateDecoration (CompWindow *w)
 {
-    Decoration *decoration;
-
     DECOR_DISPLAY (w->screen->display);
     DECOR_WINDOW (w);
 
-    decoration = decorCreateDecoration (w->screen, w->id, dd->winDecorAtom);
-
-    if (dw->decor)
-	decorReleaseDecoration (w->screen, dw->decor);
-
-    dw->decor = decoration;
+    dw->decors = decorUpdateDecorations (w->screen, w->id,
+					 dd->winDecorAtom,
+					 dw->decors, &dw->decorNum);
 }
 
 static WindowDecoration *
@@ -651,7 +857,7 @@ updateWindowDecorationScale (CompWindow *w)
 {
     WindowDecoration *wd;
     int		     x1, y1, x2, y2;
-    float            sx, sy;
+    float	     sx, sy;
     int		     i;
 
     DECOR_WINDOW (w);
@@ -717,17 +923,294 @@ decorWindowShiftY (CompWindow *w)
     return 0;
 }
 
+static void
+decorWindowUpdateFrame (CompWindow *w)
+{
+    CompDisplay	     *d = w->screen->display;
+    WindowDecoration *wd;
+
+    DECOR_DISPLAY (w->screen->display);
+    DECOR_WINDOW (w);
+
+    wd = dw->wd;
+
+    if (wd && (w->frameInput.left || w->frameInput.right ||
+	w->frameInput.top || w->frameInput.bottom ||
+	w->input.left || w->input.right || w->input.top || w->input.bottom))
+    {
+	XRectangle           rects[4];
+	int	             x, y, width, height;
+	int	             i = 0;
+	int                  bw = w->serverBorderWidth * 2;
+	CompWindowExtents    frame;
+
+	if ((w->state & MAXIMIZE_STATE) == MAXIMIZE_STATE)
+	    frame = wd->decor->maxFrame;
+	else
+	    frame = wd->decor->frame;
+
+	x      = w->frameInput.left - frame.left;
+	y      = w->frameInput.top - frame.top;
+	width  = w->serverWidth + frame.left + frame.right + bw;
+	height = w->serverHeight + frame.top  + frame.bottom + bw;
+
+	if (w->shaded)
+	    height = frame.top + frame.bottom;
+
+	if (!dw->inputFrame)
+	{
+	    XSetWindowAttributes attr;
+
+	    attr.event_mask	   = StructureNotifyMask;
+	    attr.override_redirect = TRUE;
+
+	    dw->inputFrame = XCreateWindow (d->display, w->frame,
+					    x, y, width, height, 0, CopyFromParent,
+					    InputOnly, CopyFromParent,
+					    CWOverrideRedirect | CWEventMask,
+					    &attr);
+
+	    XGrabButton (d->display, AnyButton, AnyModifier, dw->inputFrame,
+			 TRUE, ButtonPressMask | ButtonReleaseMask |
+			 ButtonMotionMask, GrabModeSync, GrabModeSync, None,
+			 None);
+
+	    XMapWindow (d->display, dw->inputFrame);
+
+	    XChangeProperty (d->display, w->id,
+			     dd->inputFrameAtom, XA_WINDOW, 32,
+			     PropModeReplace, (unsigned char *) &dw->inputFrame, 1);
+	}
+
+	XMoveResizeWindow (d->display, dw->inputFrame, x, y, width, height);
+	XLowerWindow (d->display, dw->inputFrame);
+
+	rects[i].x	= 0;
+	rects[i].y	= 0;
+	rects[i].width  = width;
+	rects[i].height = frame.top;
+
+	if (rects[i].width && rects[i].height)
+	    i++;
+
+	rects[i].x	= 0;
+	rects[i].y	= frame.top;
+	rects[i].width  = frame.left;
+	rects[i].height = height - frame.top - frame.bottom;
+
+	if (rects[i].width && rects[i].height)
+	    i++;
+
+	rects[i].x	= width - frame.right;
+	rects[i].y	= frame.top;
+	rects[i].width  = frame.right;
+	rects[i].height = height - frame.top - frame.bottom;
+
+	if (rects[i].width && rects[i].height)
+	    i++;
+
+	rects[i].x	= 0;
+	rects[i].y	= height - frame.bottom;
+	rects[i].width  = width;
+	rects[i].height = frame.bottom;
+
+	if (rects[i].width && rects[i].height)
+	    i++;
+
+	XShapeCombineRectangles (d->display, dw->inputFrame, ShapeInput,
+				 0, 0, rects, i, ShapeSet, YXBanded);
+    }
+    else
+    {
+	if (dw->inputFrame)
+	{
+	    XDeleteProperty (d->display, w->id,
+			     dd->inputFrameAtom);
+	    XDestroyWindow (d->display, dw->inputFrame);
+	    dw->inputFrame = None;
+	}
+    }
+}
+
+static Bool
+decorWindowMatchType (CompWindow *w,
+		      unsigned int decorType)
+{
+    int i;
+    unsigned int nTypeStates = 5;
+    struct typestate {
+	unsigned int compFlag;
+	unsigned int decorFlag;
+    } typeStates[] =
+    {
+	{ CompWindowTypeNormalMask, DECOR_WINDOW_TYPE_NORMAL },
+	{ CompWindowTypeDialogMask, DECOR_WINDOW_TYPE_DIALOG },
+	{ CompWindowTypeModalDialogMask, DECOR_WINDOW_TYPE_MODAL_DIALOG },
+	{ CompWindowTypeMenuMask, DECOR_WINDOW_TYPE_MENU },
+	{ CompWindowTypeUtilMask, DECOR_WINDOW_TYPE_UTILITY}
+    };
+
+    for (i = 0; i < nTypeStates; i++)
+    {
+	if ((decorType & typeStates[i].decorFlag) &&
+	    (w->type & typeStates[i].compFlag))
+	    return TRUE;
+    }
+
+    return FALSE;
+}
+
+static Bool
+decorWindowMatchState (CompWindow   *w,
+		       unsigned int decorState)
+{
+    int i;
+    unsigned int nStateStates = 3;
+    struct statestate {
+	unsigned int compFlag;
+	unsigned int decorFlag;
+    } stateStates[] =
+    {
+	{ CompWindowStateMaximizedVertMask, DECOR_WINDOW_STATE_MAXIMIZED_VERT },
+	{ CompWindowStateMaximizedHorzMask, DECOR_WINDOW_STATE_MAXIMIZED_HORZ },
+	{ CompWindowStateShadedMask, DECOR_WINDOW_STATE_SHADED }
+    };
+
+    /* active is a separate check */
+    if (w->screen->display->activeWindow == w->id)
+	decorState &= ~(DECOR_WINDOW_STATE_FOCUS);
+
+    for (i = 0; i < nStateStates; i++)
+    {
+	if ((decorState & stateStates[i].decorFlag) &&
+	    (w->state & stateStates[i].compFlag))
+	    decorState &= ~(stateStates[i].decorFlag);
+    }
+
+    return (decorState == 0);
+}
+
+static Bool
+decorWindowMatchActions (CompWindow   *w,
+			 unsigned int decorActions)
+{
+    int i;
+    unsigned int nActionStates = 16;
+    struct actionstate {
+	unsigned int compFlag;
+	unsigned int decorFlag;
+    } actionStates[] =
+    {
+	{ DECOR_WINDOW_ACTION_RESIZE_HORZ, CompWindowActionResizeMask },
+	{ DECOR_WINDOW_ACTION_RESIZE_VERT, CompWindowActionResizeMask },
+	{ DECOR_WINDOW_ACTION_CLOSE, CompWindowActionCloseMask },
+	{ DECOR_WINDOW_ACTION_MINIMIZE, CompWindowActionMinimizeMask },
+	{ DECOR_WINDOW_ACTION_UNMINIMIZE, CompWindowActionMinimizeMask },
+	{ DECOR_WINDOW_ACTION_MAXIMIZE_HORZ, CompWindowActionMaximizeHorzMask },
+	{ DECOR_WINDOW_ACTION_MAXIMIZE_VERT, CompWindowActionMaximizeVertMask },
+	{ DECOR_WINDOW_ACTION_UNMAXIMIZE_HORZ, CompWindowActionMaximizeHorzMask },
+	{ DECOR_WINDOW_ACTION_UNMAXIMIZE_VERT, CompWindowActionMaximizeVertMask },
+	{ DECOR_WINDOW_ACTION_SHADE, CompWindowActionShadeMask },
+	{ DECOR_WINDOW_ACTION_UNSHADE, CompWindowActionShadeMask },
+	{ DECOR_WINDOW_ACTION_STICK, CompWindowActionStickMask },
+	{ DECOR_WINDOW_ACTION_UNSTICK, CompWindowActionStickMask },
+	{ DECOR_WINDOW_ACTION_FULLSCREEN, CompWindowActionFullscreenMask },
+	{ DECOR_WINDOW_ACTION_ABOVE, CompWindowActionAboveMask },
+	{ DECOR_WINDOW_ACTION_BELOW, CompWindowActionBelowMask },
+    };
+
+    for (i = 0; i < nActionStates; i++)
+    {
+	if ((decorActions & actionStates[i].decorFlag) &&
+	    (w->type & actionStates[i].compFlag))
+	    decorActions &= ~(actionStates[i].decorFlag);
+    }
+
+    return (decorActions == 0);
+}
+
+static Decoration *
+decorFindMatchingDecoration (CompWindow   *w,
+			     Decoration   **decors,
+			     unsigned int decorNum,
+			     Bool	  sizeCheck)
+{
+    int	       i;
+    Decoration *decoration = NULL;
+
+    if (!decors)
+       return NULL;
+
+    if (decorNum > 0)
+    {
+	int typeMatch = (1 << 0);
+	int stateMatch = (1 << 1);
+	int actionsMatch = (1 << 2);
+
+	int currentDecorState = 0;
+
+	if (sizeCheck ? decorCheckSize (w, decors[decorNum - 1]) : TRUE)
+	    decoration = decors[decorNum - 1];
+	else
+	    decoration = NULL;
+
+	for (i = 0; i < decorNum; i++)
+	{
+	    Decoration *d = decors[i];
+
+	    /* must always match type */
+	    if (decorWindowMatchType (w, d->frameType))
+	    {
+		/* use this decoration if the type matched */
+		if (!(typeMatch & currentDecorState) &&
+		    (!sizeCheck || decorCheckSize (w, d)))
+		{
+		    decoration = d;
+		    currentDecorState |= typeMatch;
+		}
+
+		/* must always match state if type is already matched */
+		if (decorWindowMatchState (w, d->frameState) &&
+		    (!sizeCheck || decorCheckSize (w, d)))
+		{
+		    /* use this decoration if the type and state match */
+		    if (!(stateMatch & currentDecorState))
+		    {
+			decoration = d;
+			currentDecorState |= stateMatch;
+		    }
+
+		    /* must always match actions if
+		       state and type are already matched */
+		    if (decorWindowMatchActions (w, d->frameActions) &&
+			(!sizeCheck || decorCheckSize (w, d)))
+		    {
+			/* use this decoration if the requested actions match */
+			if (!(actionsMatch & currentDecorState))
+			{
+			    decoration = d;
+			    currentDecorState |= actionsMatch;
+
+			    /* perfect match, no need to continue searching */
+			    break;
+			}
+		    }
+		}
+	    }
+	}
+    }
+
+    return decoration;
+}
+
 static Bool
 decorWindowUpdate (CompWindow *w,
 		   Bool	      allowDecoration)
 {
     WindowDecoration *wd;
-    Decoration	     *old, *decor = NULL;
+    Decoration	     *old = NULL, *decoration = NULL;
     Bool	     decorate = FALSE;
     CompMatch	     *match;
-    int		     moveDx, moveDy;
-    int		     oldShiftX = 0;
-    int		     oldShiftY  = 0;
 
     DECOR_DISPLAY (w->screen->display);
     DECOR_SCREEN (w->screen);
@@ -763,64 +1246,71 @@ decorWindowUpdate (CompWindow *w,
 
     if (decorate)
     {
-	if (dw->decor && decorCheckSize (w, dw->decor))
+	/* attempt to find a matching decoration */
+	decoration = decorFindMatchingDecoration (w,
+						  dw->decors,
+						  dw->decorNum,
+						  TRUE);
+
+	if (!decoration)
 	{
-	    decor = dw->decor;
-	}
-	else
-	{
-	    if (w->id == w->screen->display->activeWindow)
-		decor = ds->decor[DECOR_ACTIVE];
-	    else
-		decor = ds->decor[DECOR_NORMAL];
+	    /* find an appropriate default decoration to use */
+	    decoration = decorFindMatchingDecoration (w,
+						      ds->decors[DECOR_ACTIVE],
+						      ds->decorNum[DECOR_ACTIVE],
+						      FALSE);
+
+	    if (!decoration)
+		decorate = FALSE;
 	}
     }
-    else
+
+    if (!decorate)
     {
 	match = &dd->opt[DECOR_DISPLAY_OPTION_SHADOW_MATCH].value.match;
 	if (matchEval (match, w))
 	{
-	    if (w->region->numRects == 1)
-		decor = ds->decor[DECOR_BARE];
+	    if (ds->decors[DECOR_BARE] && ds->decorNum[DECOR_BARE] > 0 &&
+		w->region->numRects == 1)
+		decoration = ds->decors[DECOR_BARE][0];
 
-	    if (decor)
+	    if (decoration)
 	    {
-		if (!decorCheckSize (w, decor))
-		    decor = NULL;
+		if (!decorCheckSize (w, decoration))
+		    decoration = NULL;
 	    }
 	}
     }
 
     if (!ds->dmWin || !allowDecoration)
-	decor = NULL;
+	decoration = NULL;
 
-    if (decor == old)
+    if (decoration == old)
 	return FALSE;
 
     damageWindowOutputExtents (w);
 
     if (old)
-    {
-	oldShiftX = decorWindowShiftX (w);
-	oldShiftY = decorWindowShiftY (w);
-
 	destroyWindowDecoration (w->screen, wd);
-    }
 
-    if (decor)
+    if (decoration)
     {
-	dw->wd = createWindowDecoration (decor);
+	dw->wd = createWindowDecoration (decoration);
 	if (!dw->wd)
 	    return FALSE;
 
 	if ((w->state & MAXIMIZE_STATE) == MAXIMIZE_STATE)
-	    setWindowFrameExtents (w, &decor->maxInput);
+	{
+	    setWindowFrameExtents (w, &decoration->maxFrame);
+	    setWindowBorderExtents (w, &decoration->maxBorder);
+	}
 	else
-	    setWindowFrameExtents (w, &decor->input);
-
-	moveDx = decorWindowShiftX (w) - oldShiftX;
-	moveDy = decorWindowShiftY (w) - oldShiftY;
+	{
+	    setWindowFrameExtents (w, &decoration->frame);
+	    setWindowBorderExtents (w, &decoration->border);
+	}
 
+	decorWindowUpdateFrame (w);
 	updateWindowOutputExtents (w);
 	damageWindowOutputExtents (w);
 	updateWindowDecorationScale (w);
@@ -834,35 +1324,7 @@ decorWindowUpdate (CompWindow *w,
 
 	dw->wd = NULL;
 
-	moveDx = -oldShiftX;
-	moveDy = -oldShiftY;
-    }
-
-    if (w->placed && !w->attrib.override_redirect && (moveDx || moveDy))
-    {
-	XWindowChanges xwc;
-	unsigned int   mask = CWX | CWY;
-
-	xwc.x = w->serverX + moveDx;
-	xwc.y = w->serverY + moveDy;
-
-	if (w->state & CompWindowStateFullscreenMask)
-	    mask &= ~(CWX | CWY);
-
-	if (w->state & CompWindowStateMaximizedHorzMask)
-	    mask &= ~CWX;
-
-	if (w->state & CompWindowStateMaximizedVertMask)
-	    mask &= ~CWY;
-
-	if (w->saveMask & CWX)
-	    w->saveWc.x += moveDx;
-
-	if (w->saveMask & CWY)
-	    w->saveWc.y += moveDy;
-
-	if (mask)
-	    configureXWindow (w, mask, &xwc);
+	decorWindowUpdateFrame (w);
     }
 
     return TRUE;
@@ -906,37 +1368,39 @@ decorCheckForDmOnScreen (CompScreen *s,
 	XFree (data);
     }
 
+    /* different decorator became active, update all decorations */
     if (dmWin != ds->dmWin)
     {
 	CompWindow *w;
 	int	   i;
 
+	/* create new default decorations */
 	if (dmWin)
 	{
 	    for (i = 0; i < DECOR_NUM; i++)
-		ds->decor[i] =
-		    decorCreateDecoration (s, s->root, dd->decorAtom[i]);
+		ds->decors[i] = decorUpdateDecorations (s,
+							s->root,
+							dd->decorAtom[i],
+							ds->decors[i],
+							&ds->decorNum[i]);
 	}
 	else
 	{
+	    /* no decorator active, destroy all decorations */
 	    for (i = 0; i < DECOR_NUM; i++)
 	    {
-		if (ds->decor[i])
-		{
-		    decorReleaseDecoration (s, ds->decor[i]);
-		    ds->decor[i] = 0;
-		}
+		if (ds->decors[i] && ds->decorNum[i] > 0)
+		    decorReleaseDecorations (s, ds->decors[i], &ds->decorNum[i]);
+		ds->decors[i] = NULL;
 	    }
 
 	    for (w = s->windows; w; w = w->next)
 	    {
 		DECOR_WINDOW (w);
 
-		if (dw->decor)
-		{
-		    decorReleaseDecoration (s, dw->decor);
-		    dw->decor = 0;
-		}
+		if (dw->decors && dw->decorNum > 0)
+		    decorReleaseDecorations (s, dw->decors, &dw->decorNum);
+		dw->decors = NULL;
 	    }
 	}
 
@@ -975,6 +1439,14 @@ decorHandleEvent (CompDisplay *d,
 	if (w)
 	    decorWindowUpdate (w, TRUE);
 	break;
+    case ClientMessage:
+	if (event->xclient.message_type == dd->requestFrameExtentsAtom)
+	{
+	    w = findWindowAtDisplay (d, event->xclient.window);
+	    if (w)
+		decorWindowUpdate (w, TRUE);
+	}
+	break;
     default:
 	if (event->type == d->damageEvent + XDamageNotify)
 	{
@@ -1066,12 +1538,17 @@ decorHandleEvent (CompDisplay *d,
 			{
 			    DECOR_SCREEN (s);
 
-			    if (ds->decor[i])
-				decorReleaseDecoration (s, ds->decor[i]);
+			    if (ds->decors[i] && ds->decorNum[i] > 0)
+				decorReleaseDecorations (s,
+							 ds->decors[i],
+							 &ds->decorNum[i]);
 
-			    ds->decor[i] =
-				decorCreateDecoration (s, s->root,
-						       dd->decorAtom[i]);
+			    ds->decors[i] =
+			      decorUpdateDecorations (s,
+						      s->root,
+						      dd->decorAtom[i],
+						      ds->decors[i],
+						      &ds->decorNum[i]);
 
 			    for (w = s->windows; w; w = w->next)
 				decorWindowUpdate (w, TRUE);
@@ -1081,6 +1558,29 @@ decorHandleEvent (CompDisplay *d,
 	    }
 	}
 	break;
+    case ConfigureNotify:
+	w = findTopLevelWindowAtDisplay (d, event->xproperty.window);
+	if (w)
+	{
+	    DECOR_WINDOW (w);
+	    if (dw->wd && dw->wd->decor)
+		decorWindowUpdateFrame (w);
+	}
+	break;
+    case DestroyNotify:
+	w = findTopLevelWindowAtDisplay (d, event->xproperty.window);
+	if (w)
+	{
+	    DECOR_WINDOW (w);
+	    if (dw->inputFrame &&
+		dw->inputFrame == event->xdestroywindow.window)
+	    {
+		XDeleteProperty (d->display, w->id,
+				 dd->inputFrameAtom);
+		dw->inputFrame = None;
+	    }
+	}
+	break;
     default:
 	if (d->shapeExtension && event->type == d->shapeEvent + ShapeNotify)
 	{
@@ -1111,6 +1611,26 @@ decorDamageWindowRect (CompWindow *w,
     return status;
 }
 
+static unsigned int
+decorAddSupportedAtoms (CompScreen   *s,
+			Atom         *atoms,
+			unsigned int size)
+{
+    unsigned int count;
+
+    DECOR_DISPLAY (s->display);
+    DECOR_SCREEN (s);
+
+    UNWRAP (ds, s, addSupportedAtoms);
+    count = (*s->addSupportedAtoms) (s, atoms, size);
+    WRAP (ds, s, addSupportedAtoms, decorAddSupportedAtoms);
+
+    if (count < size)
+	atoms[count++] = dd->requestFrameExtentsAtom;
+
+    return count;
+}
+
 static void
 decorGetOutputExtentsForWindow (CompWindow	  *w,
 				CompWindowExtents *output)
@@ -1330,10 +1850,36 @@ decorWindowStateChangeNotify (CompWindow   *w,
     {
 	if (dw->wd && dw->wd->decor)
 	{
+	    int oldShiftX = decorWindowShiftX (w);
+	    int oldShiftY = decorWindowShiftY (w);
+	    int moveDx, moveDy;
+
 	    if ((w->state & MAXIMIZE_STATE) == MAXIMIZE_STATE)
-		setWindowFrameExtents (w, &dw->wd->decor->maxInput);
+	    {
+		setWindowFrameExtents (w, &dw->wd->decor->maxFrame);
+		setWindowFrameExtents (w, &dw->wd->decor->maxBorder);
+	    }
 	    else
-		setWindowFrameExtents (w, &dw->wd->decor->input);
+	    {
+		setWindowFrameExtents (w, &dw->wd->decor->frame);
+		setWindowFrameExtents (w, &dw->wd->decor->border);
+	    }
+
+	    /* since we immediately update the frame extents, we must
+	       also update the stored saved window geometry in order
+	       to prevent the window from shifting back too far once
+	       unmaximized */
+
+	    moveDx = decorWindowShiftX (w) - oldShiftX;
+	    moveDy = decorWindowShiftY (w) - oldShiftY;
+
+	    if (w->saveMask & CWX)
+	        w->saveWc.x += moveDx;
+
+	    if (w->saveMask & CWY)
+	        w->saveWc.y += moveDy;
+
+	    decorWindowUpdateFrame (w);
 	}
     }
 
@@ -1501,10 +2047,12 @@ decorInitDisplay (CompPlugin  *p,
 	XInternAtom (d->display, DECOR_WINDOW_ATOM_NAME, 0);
     dd->decorAtom[DECOR_BARE] =
 	XInternAtom (d->display, DECOR_BARE_ATOM_NAME, 0);
-    dd->decorAtom[DECOR_NORMAL] =
-	XInternAtom (d->display, DECOR_NORMAL_ATOM_NAME, 0);
     dd->decorAtom[DECOR_ACTIVE] =
 	XInternAtom (d->display, DECOR_ACTIVE_ATOM_NAME, 0);
+    dd->inputFrameAtom =
+	XInternAtom (d->display, DECOR_INPUT_FRAME_ATOM_NAME, 0);
+    dd->requestFrameExtentsAtom =
+	XInternAtom (d->display, "_NET_REQUEST_FRAME_EXTENTS", 0);
 
     WRAP (dd, d, handleEvent, decorHandleEvent);
     WRAP (dd, d, matchPropertyChanged, decorMatchPropertyChanged);
@@ -1534,6 +2082,7 @@ static Bool
 decorInitScreen (CompPlugin *p,
 		 CompScreen *s)
 {
+    int         i;
     DecorScreen *ds;
 
     DECOR_DISPLAY (s->display);
@@ -1549,9 +2098,13 @@ decorInitScreen (CompPlugin *p,
 	return FALSE;
     }
 
-    memset (ds->decor, 0, sizeof (ds->decor));
+    for (i = 0; i < DECOR_NUM; i++)
+    {
+	ds->decors[i]   = NULL;
+	ds->decorNum[i] = 0;
+    }
 
-    ds->dmWin                = None;
+    ds->dmWin		     = None;
     ds->decoratorStartHandle = 0;
 
     WRAP (ds, s, drawWindow, decorDrawWindow);
@@ -1560,10 +2113,12 @@ decorInitScreen (CompPlugin *p,
     WRAP (ds, s, windowMoveNotify, decorWindowMoveNotify);
     WRAP (ds, s, windowResizeNotify, decorWindowResizeNotify);
     WRAP (ds, s, windowStateChangeNotify, decorWindowStateChangeNotify);
+    WRAP (ds, s, addSupportedAtoms, decorAddSupportedAtoms);
 
     s->base.privates[dd->screenPrivateIndex].ptr = ds;
 
     decorCheckForDmOnScreen (s, FALSE);
+    setSupportedWmHints (s);
 
     if (!ds->dmWin)
 	ds->decoratorStartHandle = compAddTimeout (0, -1,
@@ -1581,8 +2136,10 @@ decorFiniScreen (CompPlugin *p,
     DECOR_SCREEN (s);
 
     for (i = 0; i < DECOR_NUM; i++)
-	if (ds->decor[i])
-	    decorReleaseDecoration (s, ds->decor[i]);
+    {
+	if (ds->decors[i] && ds->decorNum[i] > 0)
+	    decorReleaseDecorations (s, ds->decors[i], &ds->decorNum[i]);
+    }
 
     if (ds->decoratorStartHandle)
 	compRemoveTimeout (ds->decoratorStartHandle);
@@ -1595,6 +2152,9 @@ decorFiniScreen (CompPlugin *p,
     UNWRAP (ds, s, windowMoveNotify);
     UNWRAP (ds, s, windowResizeNotify);
     UNWRAP (ds, s, windowStateChangeNotify);
+    UNWRAP (ds, s, addSupportedAtoms);
+
+    setSupportedWmHints (s);
 
     free (ds);
 }
@@ -1611,8 +2171,10 @@ decorInitWindow (CompPlugin *p,
     if (!dw)
 	return FALSE;
 
-    dw->wd    = NULL;
-    dw->decor = NULL;
+    dw->wd	   = NULL;
+    dw->decors     = NULL;
+    dw->decorNum   = 0;
+    dw->inputFrame = None;
 
     dw->resizeUpdateHandle = 0;
 
@@ -1642,8 +2204,8 @@ decorFiniWindow (CompPlugin *p,
     if (dw->wd)
 	destroyWindowDecoration (w->screen, dw->wd);
 
-    if (dw->decor)
-	decorReleaseDecoration (w->screen, dw->decor);
+    if (dw->decors && dw->decorNum > 0)
+	decorReleaseDecorations (w->screen, dw->decors, &dw->decorNum);
 
     free (dw);
 }
