@@ -74,6 +74,7 @@ typedef struct _MoveDisplay {
     int	       y;
     Region     region;
     int        status;
+    Bool       constrainY;
     KeyCode    key[NUM_KEYS];
 
     int releaseButton;
@@ -89,7 +90,10 @@ typedef struct _MoveScreen {
     Cursor moveCursor;
 
     unsigned int origState;
+    unsigned int savedWidth;
 
+    int	snapOffX;
+    int	snapBackX;
     int	snapOffY;
     int	snapBackY;
 } MoveScreen;
@@ -123,11 +127,17 @@ moveInitiate (CompDisplay     *d,
     xid = getIntOptionNamed (option, nOption, "window", 0);
 
     w = findWindowAtDisplay (d, xid);
+
+    /* make window stay above fullscreen windows while moving */
+    if (w)
+	updateWindowAttributes (w, CompStackingUpdateModeAboveFullscreen);
+
     if (w && (w->actions & CompWindowActionMoveMask))
     {
 	XRectangle   workArea;
 	unsigned int mods;
 	int          x, y, button;
+	Bool         sourceExternalApp;
 
 	MOVE_SCREEN (w->screen);
 
@@ -140,7 +150,7 @@ moveInitiate (CompDisplay     *d,
 
 	button = getIntOptionNamed (option, nOption, "button", -1);
 
-	if (otherScreenGrabExist (w->screen, "move", 0))
+	if (otherScreenGrabExist (w->screen, "move", NULL))
 	    return FALSE;
 
 	if (md->w)
@@ -171,15 +181,23 @@ moveInitiate (CompDisplay     *d,
 	md->x = 0;
 	md->y = 0;
 
+	sourceExternalApp = getBoolOptionNamed (option, nOption, "external",
+						FALSE);
+	md->constrainY = md->opt[MOVE_DISPLAY_OPTION_CONSTRAIN_Y].value.b;
+
 	lastPointerX = x;
 	lastPointerY = y;
 
 	ms->origState = w->state;
 
+	ms->savedWidth = 0;
+
 	getWorkareaForOutput (w->screen,
 			      outputDeviceForWindow (w),
 			      &workArea);
 
+	ms->snapBackX = w->serverX - workArea.x;
+	ms->snapOffX  = x - workArea.x;
 	ms->snapBackY = w->serverY - workArea.y;
 	ms->snapOffY  = y - workArea.y;
 
@@ -188,13 +206,17 @@ moveInitiate (CompDisplay     *d,
 
 	if (ms->grabIndex)
 	{
+	    unsigned int grabMask = CompWindowGrabMoveMask |
+				    CompWindowGrabButtonMask;
+
+	    if (sourceExternalApp)
+		grabMask |= CompWindowGrabExternalAppMask;
+
 	    md->w = w;
 
 	    md->releaseButton = button;
 
-	    (w->screen->windowGrabNotify) (w, x, y, mods,
-					   CompWindowGrabMoveMask |
-					   CompWindowGrabButtonMask);
+	    (w->screen->windowGrabNotify) (w, x, y, mods, grabMask);
 
 	    if (d->opt[COMP_DISPLAY_OPTION_RAISE_ON_CLICK].value.b)
 		updateWindowAttributes (w,
@@ -239,10 +261,10 @@ moveTerminate (CompDisplay     *d,
 
 	syncWindowPosition (md->w);
 
-	/* update window attributes as window constraints may have
+	/* update window size as window constraints may have
 	   changed - needed e.g. if a maximized window was moved
 	   to another output device */
-	updateWindowAttributes (md->w, CompStackingUpdateModeNone);
+	updateWindowSize (md->w);
 
 	(md->w->screen->windowUngrabNotify) (md->w);
 
@@ -355,6 +377,119 @@ moveGetYConstrainRegion (CompScreen *s)
     return region;
 }
 
+static Bool
+handleMaximizedSnapping (CompScreen *s,
+                         CompWindow *w,
+                         int xRoot, int yRoot,
+                         XRectangle workArea)
+{
+	MOVE_DISPLAY (s->display);
+	MOVE_SCREEN (s);
+
+	if (!md->opt[MOVE_DISPLAY_OPTION_SNAPOFF_MAXIMIZED].value.b)
+		return FALSE;
+
+	if (w->state & CompWindowStateMaximizedVertMask)
+	{
+		if (abs ((yRoot - workArea.y) - ms->snapOffY) >= SNAP_OFF)
+		{
+		if (!otherScreenGrabExist (s, "move", NULL))
+		{
+			int width = w->serverWidth;
+
+			w->saveMask |= CWX | CWY;
+
+			if (w->saveMask & CWWidth)
+			width = w->saveWc.width;
+
+			w->saveWc.x = xRoot - (width >> 1);
+			w->saveWc.y = yRoot + (w->input.top >> 1);
+			if (!ms->savedWidth)
+			ms->savedWidth = w->serverWidth;
+
+			md->x = md->y = 0;
+
+			maximizeWindow (w, 0);
+
+			ms->snapOffY = ms->snapBackY;
+
+			return TRUE;
+		}
+		}
+	}
+	else if (ms->origState & CompWindowStateMaximizedVertMask)
+	{
+		if (abs ((yRoot - workArea.y) - ms->snapBackY) < SNAP_BACK)
+		{
+		if (!otherScreenGrabExist (s, "move", NULL))
+		{
+			int wy;
+
+			/* update server position before maximizing
+			   window again so that it is maximized on
+			   correct output */
+			syncWindowPosition (w);
+
+			maximizeWindow (w, ms->origState);
+
+			XWindowChanges xwc;
+
+			if (ms->origState & CompWindowStateMaximizedHorzMask)
+				/* MAXIMIZE_STATE */
+				xwc.x = workArea.x;
+			else
+				/* CompWindowStateMaximizedVertMask */
+				xwc.x = xRoot - (ms->savedWidth >> 1);
+
+			xwc.width = ms->savedWidth;
+			ms->savedWidth = 0;
+			configureXWindow (w, CWX | CWWidth, &xwc);
+
+			wy  = workArea.y + (w->input.top >> 1);
+			wy += w->sizeHints.height_inc >> 1;
+
+			warpPointer (s, 0, wy - pointerY);
+
+			return TRUE;
+		}
+		}
+	}
+
+	/* Avoid handling CompWindowStateMaximizedVertMask and
+	 * CompWindowStateMaximizedHorzMask together */
+	if ((w->state & MAXIMIZE_STATE) == MAXIMIZE_STATE)
+		return FALSE;
+
+	if (w->state & CompWindowStateMaximizedHorzMask)
+	{
+		if (abs ((xRoot - workArea.x) - ms->snapOffX) >= SNAP_OFF)
+		{
+		if (!otherScreenGrabExist (s, "move", NULL))
+		{
+			int width = w->serverWidth;
+
+			w->saveMask |= CWX | CWY;
+
+			if (w->saveMask & CWWidth)
+			width = w->saveWc.width;
+
+			w->saveWc.x = xRoot - (width >> 1);
+			w->saveWc.y = yRoot + (w->input.top >> 1);
+
+			md->x = md->y = 0;
+
+			maximizeWindow (w, 0);
+
+			ms->snapOffX = ms->snapBackX;
+
+			return TRUE;
+		}
+		}
+	}
+
+	return FALSE;
+}
+
 static void
 moveHandleMotionEvent (CompScreen *s,
 		       int	  xRoot,
@@ -397,7 +532,7 @@ moveHandleMotionEvent (CompScreen *s,
 				  outputDeviceForWindow (w),
 				  &workArea);
 
-	    if (md->opt[MOVE_DISPLAY_OPTION_CONSTRAIN_Y].value.b)
+	    if (md->constrainY)
 	    {
 		if (!md->region)
 		    md->region = moveGetYConstrainRegion (s);
@@ -410,8 +545,8 @@ moveHandleMotionEvent (CompScreen *s,
 		    int x, y, width, height;
 		    int status;
 
-		    x	   = wX + dx - w->input.left;
-		    y	   = wY + dy - w->input.top;
+		    x	   = (wX + dx - w->input.left) + w->clientFrame.left;
+		    y	   = (wY + dy - w->input.top) + w->clientFrame.top;
 		    width  = wWidth + w->input.left + w->input.right;
 		    height = w->input.top ? w->input.top : 1;
 
@@ -453,59 +588,8 @@ moveHandleMotionEvent (CompScreen *s,
 		}
 	    }
 
-	    if (md->opt[MOVE_DISPLAY_OPTION_SNAPOFF_MAXIMIZED].value.b)
-	    {
-		if (w->state & CompWindowStateMaximizedVertMask)
-		{
-		    if (abs ((yRoot - workArea.y) - ms->snapOffY) >= SNAP_OFF)
-		    {
-			if (!otherScreenGrabExist (s, "move", 0))
-			{
-			    int width = w->serverWidth;
-
-			    w->saveMask |= CWX | CWY;
-
-			    if (w->saveMask & CWWidth)
-				width = w->saveWc.width;
-
-			    w->saveWc.x = xRoot - (width >> 1);
-			    w->saveWc.y = yRoot + (w->input.top >> 1);
-
-			    md->x = md->y = 0;
-
-			    maximizeWindow (w, 0);
-
-			    ms->snapOffY = ms->snapBackY;
-
-			    return;
-			}
-		    }
-		}
-		else if (ms->origState & CompWindowStateMaximizedVertMask)
-		{
-		    if (abs ((yRoot - workArea.y) - ms->snapBackY) < SNAP_BACK)
-		    {
-			if (!otherScreenGrabExist (s, "move", 0))
-			{
-			    int wy;
-
-			    /* update server position before maximizing
-			       window again so that it is maximized on
-			       correct output */
-			    syncWindowPosition (w);
-
-			    maximizeWindow (w, ms->origState);
-
-			    wy  = workArea.y + (w->input.top >> 1);
-			    wy += w->sizeHints.height_inc >> 1;
-
-			    warpPointer (s, 0, wy - pointerY);
-
-			    return;
-			}
-		    }
-		}
-	    }
+	    if (handleMaximizedSnapping(s, w, xRoot, yRoot, workArea))
+		    return;
 
 	    if (w->state & CompWindowStateMaximizedVertMask)
 	    {
@@ -638,7 +722,7 @@ moveHandleEvent (CompDisplay *d,
 		w = findWindowAtDisplay (d, event->xclient.window);
 		if (w)
 		{
-		    CompOption o[5];
+		    CompOption o[6];
 		    int	       xRoot, yRoot;
 		    int	       option;
 
@@ -646,13 +730,17 @@ moveHandleEvent (CompDisplay *d,
 		    o[0].name    = "window";
 		    o[0].value.i = event->xclient.window;
 
+		    o[1].type    = CompOptionTypeBool;
+		    o[1].name    = "external";
+		    o[1].value.b = TRUE;
+
 		    if (event->xclient.data.l[2] == WmMoveResizeMoveKeyboard)
 		    {
 			option = MOVE_DISPLAY_OPTION_INITIATE_KEY;
 
 			moveInitiate (d, &md->opt[option].value.action,
 				      CompActionStateInitKey,
-				      o, 1);
+				      o, 2);
 		    }
 		    else
 		    {
@@ -669,27 +757,27 @@ moveHandleEvent (CompDisplay *d,
 			/* TODO: not only button 1 */
 			if (mods & Button1Mask)
 			{
-			    o[1].type	 = CompOptionTypeInt;
-			    o[1].name	 = "modifiers";
-			    o[1].value.i = mods;
-
 			    o[2].type	 = CompOptionTypeInt;
-			    o[2].name	 = "x";
-			    o[2].value.i = event->xclient.data.l[0];
+			    o[2].name	 = "modifiers";
+			    o[2].value.i = mods;
 
 			    o[3].type	 = CompOptionTypeInt;
-			    o[3].name	 = "y";
-			    o[3].value.i = event->xclient.data.l[1];
+			    o[3].name	 = "x";
+			    o[3].value.i = event->xclient.data.l[0];
 
-			    o[4].type    = CompOptionTypeInt;
-			    o[4].name    = "button";
-			    o[4].value.i = event->xclient.data.l[3] ?
+			    o[4].type	 = CompOptionTypeInt;
+			    o[4].name	 = "y";
+			    o[4].value.i = event->xclient.data.l[1];
+
+			    o[5].type    = CompOptionTypeInt;
+			    o[5].name    = "button";
+			    o[5].value.i = event->xclient.data.l[3] ?
 				           event->xclient.data.l[3] : -1;
 
 			    moveInitiate (d,
 					  &md->opt[option].value.action,
 					  CompActionStateInitButton,
-					  o, 5);
+					  o, 6);
 
 			    moveHandleMotionEvent (w->screen, xRoot, yRoot);
 			}
@@ -875,6 +963,7 @@ moveInitDisplay (CompPlugin  *p,
     md->region        = NULL;
     md->status        = RectangleOut;
     md->releaseButton = 0;
+    md->constrainY    = FALSE;
 
     for (i = 0; i < NUM_KEYS; i++)
 	md->key[i] = XKeysymToKeycode (d->display,
@@ -899,6 +988,9 @@ moveFiniDisplay (CompPlugin  *p,
 
     compFiniDisplayOptions (d, md->opt, MOVE_DISPLAY_OPTION_NUM);
 
+    if (md->region)
+	XDestroyRegion (md->region);
+
     free (md);
 }
 
