@@ -16,8 +16,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
+ * 02110-1301, USA
  */
 
 #include <math.h>
@@ -131,12 +131,15 @@ typedef enum {
 
 /* helper macros to get the full dimensions of a window,
    including decorations */
+#define BORDER_WIDTH(w)  ((w)->input.left + (w)->input.right + \
+			  2 * (w)->serverBorderWidth)
+#define BORDER_HEIGHT(w) ((w)->input.top + (w)->input.bottom + \
+			  2 * (w)->serverBorderWidth)
+
 #define WIN_FULL_X(w) ((w)->serverX - (w)->input.left)
 #define WIN_FULL_Y(w) ((w)->serverY - (w)->input.top)
-#define WIN_FULL_W(w) ((w)->serverWidth + 2 * (w)->serverBorderWidth + \
-		       (w)->input.left + (w)->input.right)
-#define WIN_FULL_H(w) ((w)->serverHeight + 2 * (w)->serverBorderWidth + \
-		       (w)->input.top + (w)->input.bottom)
+#define WIN_FULL_W(w) ((w)->serverWidth + BORDER_WIDTH (w))
+#define WIN_FULL_H(w) ((w)->serverHeight + BORDER_HEIGHT (w))
 
 static Bool
 placeMatchXYValue (CompWindow *w,
@@ -844,8 +847,8 @@ placePointer (CompWindow *w,
 
     if (placeGetPointerPosition (w->screen, &xPointer, &yPointer))
     {
-	*x = xPointer - (w->serverWidth / 2);
-	*y = yPointer - (w->serverHeight / 2);
+	*x = xPointer - (w->serverWidth / 2) - w->serverBorderWidth;
+	*y = yPointer - (w->serverHeight / 2) - w->serverBorderWidth;
     }
     else
     {
@@ -1042,13 +1045,38 @@ placeSmart (CompWindow *w,
     *y = yOptimal + w->input.top;
 }
 
-static PlacementStrategy
-placeGetStrategyForWindow (CompWindow *w)
+static Bool
+placeWindowHasUserDefinedPosition (CompWindow *w,
+				   Bool       acceptPPosition)
 {
     CompMatch *match;
 
     PLACE_SCREEN (w->screen);
 
+    match = &ps->opt[PLACE_SCREEN_OPTION_FORCE_PLACEMENT].value.match;
+    if (matchEval (match, w))
+	return FALSE;
+
+    if (acceptPPosition && (w->sizeHints.flags & PPosition))
+	return TRUE;
+
+    if ((w->type & CompWindowTypeNormalMask) ||
+	ps->opt[PLACE_SCREEN_OPTION_WORKAROUND].value.b)
+    {
+	/* Only accept USPosition on non-normal windows if workarounds are
+	 * enabled because apps claiming the user set -geometry for a
+	 * dialog or dock are most likely wrong
+	 */
+	if (w->sizeHints.flags & USPosition)
+	    return TRUE;
+    }
+
+    return FALSE;
+}
+
+static PlacementStrategy
+placeGetStrategyForWindow (CompWindow *w)
+{
     if (w->type & (CompWindowTypeDockMask | CompWindowTypeDesktopMask    |
 		   CompWindowTypeUtilMask | CompWindowTypeToolbarMask    |
 		   CompWindowTypeMenuMask | CompWindowTypeFullscreenMask |
@@ -1064,29 +1092,10 @@ placeGetStrategyForWindow (CompWindow *w)
 	return NoPlacement;
     }
 
-    /* no placement for unmovable windows */
-    if (!(w->actions & CompWindowActionMoveMask))
-	return NoPlacement;
+    if (placeWindowHasUserDefinedPosition (w, TRUE))
+	return ConstrainOnly;
 
-    match = &ps->opt[PLACE_SCREEN_OPTION_FORCE_PLACEMENT].value.match;
-    if (!matchEval (match, w))
-    {
-	if ((w->type & CompWindowTypeNormalMask) ||
-	    ps->opt[PLACE_SCREEN_OPTION_WORKAROUND].value.b)
-	{
-	    /* Only accept USPosition on non-normal windows if workarounds are
-	     * enabled because apps claiming the user set -geometry for a
-	     * dialog or dock are most likely wrong
-	     */
-	    if (w->sizeHints.flags & USPosition)
-		return ConstrainOnly;
-	}
-
-	if (w->sizeHints.flags & PPosition)
-	    return ConstrainOnly;
-    }
-
-   if (w->transientFor &&
+    if (w->transientFor &&
 	(w->type & (CompWindowTypeDialogMask |
 		    CompWindowTypeModalDialogMask)))
     {
@@ -1216,8 +1225,8 @@ placeConstrainToWorkarea (CompWindow *w,
 
     extents.left   = *x - w->input.left;
     extents.top    = *y - w->input.top;
-    extents.right  = *x + w->serverWidth + w->input.right;
-    extents.bottom = *y + w->serverHeight + w->input.bottom;
+    extents.right  = extents.left + WIN_FULL_W (w);
+    extents.bottom = extents.top + WIN_FULL_H (w);
 
     delta = workArea->x + workArea->width - extents.right;
     if (delta < 0)
@@ -1419,9 +1428,9 @@ placeDoValidateWindowResizeRequest (CompWindow     *w,
     }
 
     left   = x - w->input.left;
-    right  = x + xwc->width + w->input.right;
+    right  = left + xwc->width + BORDER_WIDTH (w);
     top    = y - w->input.top;
-    bottom = y + xwc->height + w->input.bottom;
+    bottom = top + xwc->height + BORDER_HEIGHT (w);
 
     output = outputDeviceForGeometry (s,
 				      xwc->x, xwc->y,
@@ -1484,9 +1493,14 @@ placeDoValidateWindowResizeRequest (CompWindow     *w,
 
     /* bring left/right/top/bottom to actual window coordinates */
     left   += w->input.left;
-    right  -= w->input.right;
+    right  -= w->input.right + 2 * w->serverBorderWidth;
     top    += w->input.top;
-    bottom -= w->input.bottom;
+    bottom -= w->input.bottom + 2 * w->serverBorderWidth;
+
+    /* always validate position if the application changed only its size,
+       as it might become partially offscreen because of that */
+    if (!(*mask & (CWX | CWY)) && (*mask & (CWWidth | CWHeight)))
+	sizeOnly = FALSE;
 
     if ((right - left) != xwc->width)
     {
@@ -1549,17 +1563,20 @@ placeValidateWindowResizeRequest (CompWindow     *w,
 		     CompWindowTypeDesktopMask))
 	return;
 
-    if (w->sizeHints.flags & USPosition)
+    /* do nothing if the window was already (at least partially) offscreen */
+    if (w->serverX < 0                         ||
+	w->serverX + w->serverWidth > s->width ||
+	w->serverY < 0                         ||
+	w->serverY + w->serverHeight > s->height)
     {
-	/* only respect USPosition on normal windows if
-	   workarounds are disabled, reason see above */
-	if (ps->opt[PLACE_SCREEN_OPTION_WORKAROUND].value.b ||
-	    (w->type & CompWindowTypeNormalMask))
-	{
-	    /* try to keep the window position intact for USPosition -
-	       obviously we can't do that if we need to change the size */
-	    sizeOnly = TRUE;
-	}
+	return;
+    }
+
+    if (placeWindowHasUserDefinedPosition (w, FALSE))
+    {
+	/* try to keep the window position intact for USPosition -
+	   obviously we can't do that if we need to change the size */
+	sizeOnly = TRUE;
     }
 
     placeDoValidateWindowResizeRequest (w, mask, xwc, sizeOnly, TRUE);
@@ -1877,17 +1894,13 @@ placeDoHandleScreenSizeChange (CompScreen *s,
 		{
 		    mask |= CWX | CWWidth;
 		    xwc.x = vpX * s->width + workArea.x + w->input.left;
-		    xwc.width = workArea.width -
-			(2 * w->serverBorderWidth +
-			 w->input.left + w->input.right);
+		    xwc.width = workArea.width - BORDER_WIDTH (w);
 		}
 		if (w->state & CompWindowStateMaximizedVertMask)
 		{
 		    mask |= CWY | CWHeight;
 		    xwc.y = vpY * s->height + workArea.y + w->input.top;
-		    xwc.height = workArea.height -
-			(2 * w->serverBorderWidth +
-			 w->input.top + w->input.bottom);
+		    xwc.height = workArea.height - BORDER_HEIGHT (w);
 		}
 	    }
 	}
