@@ -100,7 +100,11 @@ typedef struct _ResizeDisplay {
     int		 pointerDy;
     KeyCode	 key[NUM_KEYS];
 
-    Bool offWorkAreaConstrained;
+    Region       constraintRegion;
+    int          inRegionStatus;
+    int          lastGoodHotSpotY;
+    int          lastGoodWidth;
+    int          lastGoodHeight;
 } ResizeDisplay;
 
 typedef struct _ResizeScreen {
@@ -121,8 +125,6 @@ typedef struct _ResizeScreen {
     Cursor downRightCursor;
     Cursor middleCursor;
     Cursor cursor[NUM_KEYS];
-
-    const XRectangle *grabWindowWorkArea;
 } ResizeScreen;
 
 #define GET_RESIZE_DISPLAY(d)					    \
@@ -301,6 +303,7 @@ static void
 resizeFinishResizing (CompDisplay *d)
 {
     RESIZE_DISPLAY (d);
+    BoxRec box;
 
     (*rd->w->screen->windowUngrabNotify) (rd->w);
 
@@ -308,9 +311,51 @@ resizeFinishResizing (CompDisplay *d)
 		     rd->w->id,
 		     rd->resizeInformationAtom);
 
+    if (rd->mode != RESIZE_MODE_NORMAL) {
+	if (rd->mode == RESIZE_MODE_STRETCH)
+	    resizeGetStretchRectangle (d, &box);
+	else
+	    resizeGetPaintRectangle (d, &box);
+	resizeDamageRectangle (rd->w->screen, &box);
+    }
+
     rd->w = NULL;
 }
 
+static Region
+resizeGetConstraintRegion (CompWindow *w, CompScreen *s)
+{
+    Region       region;
+    int          i , TempX, TempY, TempWidth, TempHeight;
+
+    region = XCreateRegion ();
+    if (!region)
+	return NULL;
+    for (i = 0; i < s->nOutputDev; i++){
+        /*Save the original values here*/
+        TempX = s->outputDev[i].workArea.x;
+        TempY = s->outputDev[i].workArea.y;
+        TempWidth = s->outputDev[i].workArea.width;
+        TempHeight = s->outputDev[i].workArea.height;
+        /*Allow for GtkFrameExtents*/
+        s->outputDev[i].workArea.x = s->outputDev[i].workArea.x - w->clientFrame.left;
+        s->outputDev[i].workArea.width = s->outputDev[i].workArea.width
+                                         + (w->clientFrame.left + w->clientFrame.right);
+        s->outputDev[i].workArea.y = s->outputDev[i].workArea.y - w->clientFrame.top;
+        s->outputDev[i].workArea.height = s->outputDev[i].workArea.height
+                                        + (w->clientFrame.top + w->clientFrame.bottom);
+
+	    XUnionRectWithRegion (&s->outputDev[i].workArea, region, region);
+
+         /*Reset original workarea values*/
+        s->outputDev[i].workArea.x = TempX;
+        s->outputDev[i].workArea.y = TempY;
+        s->outputDev[i].workArea.width = TempWidth;
+        s->outputDev[i].workArea.height = TempHeight;
+        }
+    return region;
+}
+
 static Bool
 resizeInitiate (CompDisplay     *d,
 		CompAction      *action,
@@ -493,18 +538,24 @@ resizeInitiate (CompDisplay     *d,
 		warpPointer (w->screen, xRoot - pointerX, yRoot - pointerY);
 	    }
 
-	    /* Update offWorkAreaConstrained and workArea at grab time */
-	    rd->offWorkAreaConstrained = FALSE;
+	    if (rd->constraintRegion)
+		XDestroyRegion (rd->constraintRegion);
+
 	    if (sourceExternalApp)
 	    {
-		int output = outputDeviceForWindow (w);
-
-		rs->grabWindowWorkArea = &w->screen->outputDev[output].workArea;
-
 		/* Prevent resizing beyond work area edges when resize is
 		   initiated externally (e.g. with window frame or menu)
 		   and not with a key (e.g. alt+button) */
-		rd->offWorkAreaConstrained = TRUE;
+
+		rd->inRegionStatus   = RectangleOut;
+		rd->lastGoodHotSpotY = -1;
+		rd->lastGoodWidth    = w->serverWidth;
+		rd->lastGoodHeight   = w->serverHeight;
+		rd->constraintRegion = resizeGetConstraintRegion (w, w->screen);
+	    }
+	    else
+	    {
+		rd->constraintRegion = NULL;
 	    }
 	}
     }
@@ -550,18 +601,7 @@ resizeTerminate (CompDisplay	 *d,
 	    else
 		geometry = rd->geometry;
 
-	    if (memcmp (&geometry, &rd->savedGeometry, sizeof (geometry)) == 0)
-	    {
-		BoxRec box;
-
-		if (rd->mode == RESIZE_MODE_STRETCH)
-		    resizeGetStretchRectangle (d, &box);
-		else
-		    resizeGetPaintRectangle (d, &box);
-
-		resizeDamageRectangle (w->screen, &box);
-	    }
-	    else
+	    if (!memcmp (&geometry, &rd->savedGeometry, sizeof (geometry)) == 0)
 	    {
 		xwc.x      = geometry.x;
 		xwc.y      = geometry.y;
@@ -586,6 +626,20 @@ resizeTerminate (CompDisplay	 *d,
 	    configureXWindow (w, mask, &xwc);
 	}
 
+	/*
+	 * XXX: This is just wacko. Only terminate if we DIDN'T change,
+	 * because if we did change, we'll shortly get a ResizeNotify about
+	 * it at which point we WILL terminate. Weird and buggy.
+	 *
+	 * It actually means we might paint resize-logic AFTER terminating
+	 * the resize, since we could get a paint-event before we get the
+	 * resize-one.
+	 *
+	 * I don't currently know why it was done like this and I'm
+	 * reluctant to just remove the if() clause until I do.
+	 *
+	 * - Kristian
+	 */
 	if (!(mask & (CWWidth | CWHeight)))
 	    resizeFinishResizing (d);
 
@@ -694,7 +748,10 @@ resizeHandleMotionEvent (CompScreen *s,
     if (rs->grabIndex)
     {
 	BoxRec box;
-	int    w, h;
+	int    w, h;                    /* size of window contents */
+	int    wX, wY, wWidth, wHeight; /* rect. for window contents+borders */
+	int    i;
+	int    workAreaSnapDistance = 15;
 
 	RESIZE_DISPLAY (s->display);
 
@@ -703,16 +760,16 @@ resizeHandleMotionEvent (CompScreen *s,
 
 	if (!rd->mask)
 	{
-	    CompWindow *w = rd->w;
+	    CompWindow *cw = rd->w;
 	    int        xDist, yDist;
 	    int        minPointerOffsetX, minPointerOffsetY;
 
-	    xDist = xRoot - (w->serverX + (w->serverWidth / 2));
-	    yDist = yRoot - (w->serverY + (w->serverHeight / 2));
+	    xDist = xRoot - (cw->serverX + (cw->serverWidth / 2));
+	    yDist = yRoot - (cw->serverY + (cw->serverHeight / 2));
 
 	    /* decision threshold is 10% of window size */
-	    minPointerOffsetX = MIN (20, w->serverWidth / 10);
-	    minPointerOffsetY = MIN (20, w->serverHeight / 10);
+	    minPointerOffsetX = MIN (20, cw->serverWidth / 10);
+	    minPointerOffsetY = MIN (20, cw->serverHeight / 10);
 
 	    /* if we reached the threshold in one direction,
 	       make the threshold in the other direction smaller
@@ -757,16 +814,16 @@ resizeHandleMotionEvent (CompScreen *s,
 		action->state |= CompActionStateTermButton;
 
 		if (rd->mask & ResizeRightMask)
-			pointerAdjustX = w->serverX + w->serverWidth +
-					 w->input.right - xRoot;
+			pointerAdjustX = cw->serverX + cw->serverWidth +
+					 cw->input.right - xRoot;
 		else if (rd->mask & ResizeLeftMask)
-			pointerAdjustX = w->serverX - w->input.left - xRoot;
+			pointerAdjustX = cw->serverX - cw->input.left - xRoot;
 
 		if (rd->mask & ResizeDownMask)
-			pointerAdjustY = w->serverY + w->serverHeight +
-					 w->input.bottom - yRoot;
+			pointerAdjustY = cw->serverY + cw->serverHeight +
+					 cw->input.bottom - yRoot;
 		else if (rd->mask & ResizeUpMask)
-			pointerAdjustY = w->serverY - w->input.top - yRoot;
+			pointerAdjustY = cw->serverY - cw->input.top - yRoot;
 
 		warpPointer (s, pointerAdjustX, pointerAdjustY);
 
@@ -801,42 +858,235 @@ resizeHandleMotionEvent (CompScreen *s,
 
 	constrainNewWindowSize (rd->w, w, h, &w, &h);
 
-	/* constrain to work area */
-	if (rd->offWorkAreaConstrained)
+	/* compute rect. for window + borders */
+	wWidth  = w + rd->w->input.left + rd->w->input.right;
+	wHeight = h + rd->w->input.top + rd->w->input.bottom;
+
+	if (rd->mask & ResizeLeftMask)
+	    wX = rd->savedGeometry.x + rd->savedGeometry.width -
+		 (w + rd->w->input.left);
+	else
+	    wX = rd->savedGeometry.x - rd->w->input.left;
+
+	if (rd->mask & ResizeUpMask)
+	    wY = rd->savedGeometry.y + rd->savedGeometry.height -
+		 (h + rd->w->input.top);
+	else
+	    wY = rd->savedGeometry.y - rd->w->input.top;
+
+	/* Check if resized edge(s) are near a work-area boundary */
+	for (i = 0; i < s->nOutputDev; i++)
 	{
-	    if (rd->mask & ResizeUpMask)
-	    {
-		int decorTop = rd->savedGeometry.y + rd->savedGeometry.height -
-		    (h + rd->w->input.top);
+	    const XRectangle *workArea = &s->outputDev[i].workArea;
 
-		if (rs->grabWindowWorkArea->y > decorTop)
-		    h -= rs->grabWindowWorkArea->y - decorTop;
-	    }
-	    if (rd->mask & ResizeDownMask)
+	    /* if window and work-area intersect in x axis */
+	    if (wX + wWidth > workArea->x &&
+		wX < workArea->x + workArea->width)
 	    {
-		int decorBottom = rd->savedGeometry.y + h + rd->w->input.bottom;
+		if (rd->mask & ResizeLeftMask)
+		{
+		    int dw = workArea->x - wX;
 
-		if (decorBottom >
-		    rs->grabWindowWorkArea->y + rs->grabWindowWorkArea->height)
-		    h -= decorBottom - (rs->grabWindowWorkArea->y +
-					rs->grabWindowWorkArea->height);
+		    if (0 < dw && dw < workAreaSnapDistance)
+		    {
+			w      -= dw;
+			wWidth -= dw;
+			wX     += dw;
+		    }
+		}
+		else if (rd->mask & ResizeRightMask)
+		{
+		    int dw = wX + wWidth - (workArea->x + workArea->width);
+
+		    if (0 < dw && dw < workAreaSnapDistance)
+		    {
+			w      -= dw;
+			wWidth -= dw;
+		    }
+		}
 	    }
+
+	    /* if window and work-area intersect in y axis */
+	    if (wY + wHeight > workArea->y &&
+		wY < workArea->y + workArea->height)
+	    {
+		if (rd->mask & ResizeUpMask)
+		{
+		    int dh = workArea->y - wY;
+
+		    if (0 < dh && dh < workAreaSnapDistance)
+		    {
+			h       -= dh;
+			wHeight -= dh;
+			wY      += dh;
+		    }
+		}
+		else if (rd->mask & ResizeDownMask)
+		{
+		    int dh = wY + wHeight - (workArea->y + workArea->height);
+
+		    if (0 < dh && dh < workAreaSnapDistance)
+		    {
+			h       -= dh;
+			wHeight -= dh;
+		    }
+		}
+	    }
+	}
+
+	if (rd->constraintRegion)
+	{
+	    int minWidth  = 50;
+	    int minHeight = 50;
+
+	    /* rect. for a minimal height window + borders
+	       (used for the constraining in X axis) */
+	    int minimalInputHeight = minHeight +
+		                     rd->w->input.top + rd->w->input.bottom;
+
+	    /* small hot-spot square (on window's corner or edge) that is to be
+	       constrained to the combined output work-area region */
+	    int x, y;
+	    int width = rd->w->input.top; /* square size = title bar height */
+	    int height = width;
+	    int status;
+
+	    /* compute x & y for constrained hot-spot rect */
 	    if (rd->mask & ResizeLeftMask)
-	    {
-		int decorLeft = rd->savedGeometry.x + rd->savedGeometry.width -
-		    (w + rd->w->input.left);
+		x = wX;
+	    else if (rd->mask & ResizeRightMask)
+		x = wX + wWidth - width;
+	    else
+		x = MIN (MAX (xRoot, wX), wX + wWidth - width);
 
-		if (rs->grabWindowWorkArea->x > decorLeft)
-		    w -= rs->grabWindowWorkArea->x - decorLeft;
+	    if (rd->mask & ResizeUpMask)
+		y = wY;
+	    else if (rd->mask & ResizeDownMask)
+		y = wY + wHeight - height;
+	    else
+		y = MIN (MAX (yRoot, wY), wY + wHeight - height);
+
+	    status = XRectInRegion (rd->constraintRegion,
+				    x, y, width, height);
+
+	    /* only constrain movement if previous position was valid */
+	    if (rd->inRegionStatus == RectangleIn)
+	    {
+		int xStatus=-1, yForXResize;
+		int nx = x;
+		int nw = w;
+		int nh = h;
+
+		if (rd->mask & (ResizeLeftMask | ResizeRightMask))
+		{
+		    xStatus = status;
+
+		    if (rd->mask & ResizeUpMask)
+			yForXResize = wY + wHeight - minimalInputHeight;
+		    else if (rd->mask & ResizeDownMask)
+			yForXResize = wY + minimalInputHeight - height;
+		    else
+			yForXResize = y;
+
+		    if (XRectInRegion (rd->constraintRegion,
+				       x, yForXResize,
+				       width, height) != RectangleIn)
+		    {
+			if (rd->lastGoodHotSpotY >= 0)
+			    yForXResize = rd->lastGoodHotSpotY;
+			else
+			    yForXResize = y;
+		    }
+		    if (rd->mask & ResizeLeftMask)
+		    {
+			while ((nw > minWidth) && xStatus != RectangleIn)
+			{
+			    xStatus = XRectInRegion (rd->constraintRegion,
+						     nx, yForXResize, width, height);
+			    if (xStatus != RectangleIn)
+			    {
+				nw--;
+				nx++;
+			    }
+			}
+			if (nw > minWidth)
+			{
+			    x = nx;
+			    w = nw;
+			}
+		    }
+		    else if (rd->mask & ResizeRightMask)
+		    {
+			while ((nw > minWidth) && xStatus != RectangleIn)
+			{
+			    xStatus = XRectInRegion (rd->constraintRegion,
+						     nx, yForXResize,
+						     width, height);
+			    if (xStatus != RectangleIn)
+			    {
+				nw--;
+				nx--;
+			    }
+			}
+			if (nw > minWidth)
+			{
+			    x = nx;
+			    w = nw;
+			}
+		    }
+		}
+
+		if (rd->mask & ResizeUpMask)
+		{
+		    while ((nh > minHeight) && status != RectangleIn)
+		    {
+			status = XRectInRegion (rd->constraintRegion,
+						x, y, width, height);
+			if (status != RectangleIn)
+			{
+			    nh--;
+			    y++;
+			}
+		    }
+		    if (nh > minHeight)
+			h = nh;
+		}
+		else if (rd->mask & ResizeDownMask)
+		{
+		    while ((nh > minHeight) && status != RectangleIn)
+		    {
+			status = XRectInRegion (rd->constraintRegion,
+						x, y, width, height);
+			if (status != RectangleIn)
+			{
+			    nh--;
+			    y--;
+			}
+		    }
+		    if (nh > minHeight)
+			h = nh;
+		}
+
+		if (((rd->mask & (ResizeLeftMask | ResizeRightMask)) &&
+		     xStatus == RectangleIn) ||
+		    ((rd->mask & (ResizeUpMask | ResizeDownMask)) &&
+		     status == RectangleIn))
+		{
+		    /* hot-spot inside work-area region, store good values */
+		    rd->lastGoodHotSpotY = y;
+		    rd->lastGoodWidth    = w;
+		    rd->lastGoodHeight   = h;
+		}
+		else
+		{
+		    /* failed to find a good hot-spot position, restore size */
+		    w = rd->lastGoodWidth;
+		    h = rd->lastGoodHeight;
+		}
 	    }
-	    if (rd->mask & ResizeRightMask)
+	    else
 	    {
-		int decorRight = rd->savedGeometry.x + w + rd->w->input.right;
-
-		if (decorRight >
-		    rs->grabWindowWorkArea->x + rs->grabWindowWorkArea->width)
-		    w -= decorRight - (rs->grabWindowWorkArea->x +
-				       rs->grabWindowWorkArea->width);
+		rd->inRegionStatus = status;
 	    }
 	}
 
@@ -1096,6 +1346,7 @@ resizePaintRectangle (CompScreen              *s,
 {
     BoxRec        box;
     CompTransform sTransform = *transform;
+    GLint srcAlpha, dstAlpha;
 
     resizeGetPaintRectangle (s->display, &box);
 
@@ -1107,6 +1358,9 @@ resizePaintRectangle (CompScreen              *s,
 
     glDisableClientState (GL_TEXTURE_COORD_ARRAY);
     glEnable (GL_BLEND);
+    glGetIntegerv(GL_BLEND_SRC_ALPHA, &srcAlpha);
+    glGetIntegerv(GL_BLEND_DST_ALPHA, &dstAlpha);
+    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 
     /* fill rectangle */
     if (fillColor)
@@ -1127,6 +1381,7 @@ resizePaintRectangle (CompScreen              *s,
 
     /* clean up */
     glColor4usv (defaultColor);
+    glBlendFunc (srcAlpha, dstAlpha);
     glDisable (GL_BLEND);
     glEnableClientState (GL_TEXTURE_COORD_ARRAY);
     glPopMatrix ();
@@ -1357,7 +1612,7 @@ resizeInitDisplay (CompPlugin  *p,
 	rd->key[i] = XKeysymToKeycode (d->display,
 				       XStringToKeysym (rKeys[i].name));
 
-    rd->offWorkAreaConstrained = TRUE;
+    rd->constraintRegion = NULL;
 
     WRAP (rd, d, handleEvent, resizeHandleEvent);
 
@@ -1378,6 +1633,9 @@ resizeFiniDisplay (CompPlugin  *p,
 
     compFiniDisplayOptions (d, rd->opt, RESIZE_DISPLAY_OPTION_NUM);
 
+    if (rd->constraintRegion)
+	XDestroyRegion (rd->constraintRegion);
+
     free (rd);
 }
 
@@ -1416,8 +1674,6 @@ resizeInitScreen (CompPlugin *p,
     rs->cursor[2] = rs->upCursor;
     rs->cursor[3] = rs->downCursor;
 
-    rs->grabWindowWorkArea = NULL;
-
     WRAP (rs, s, windowResizeNotify, resizeWindowResizeNotify);
     WRAP (rs, s, paintOutput, resizePaintOutput);
     WRAP (rs, s, paintWindow, resizePaintWindow);
