@@ -41,6 +41,11 @@
 #include <X11/extensions/Xcomposite.h>
 #include <X11/extensions/Xrandr.h>
 #include <X11/extensions/shape.h>
+#ifdef HAVE_XINPUT2
+#include <X11/extensions/XInput2.h>
+#endif
+#include <X11/Xcursor/Xcursor.h>
+#include <X11/cursorfont.h>
 
 #include <compiz-core.h>
 
@@ -474,6 +479,43 @@ shade (CompDisplay     *d,
     return TRUE;
 }
 
+static void
+compDisplaySetCursorTheme (CompDisplay *display)
+{
+    const char *theme = display->opt[COMP_DISPLAY_OPTION_CURSOR_THEME].value.s;
+    int size = display->opt[COMP_DISPLAY_OPTION_CURSOR_SIZE].value.i;
+
+    if (theme && strlen(theme))
+    {
+	CompScreen *s;
+#ifdef HAVE_XINPUT2
+	int clientPointer;
+
+	XIGetClientPointer (display->display, None, &clientPointer);
+#endif
+
+	XcursorSetTheme (display->display, theme);
+	if (size > 0)
+	    XcursorSetDefaultSize (display->display, size);
+
+	for (s = display->screens; s; s = s->next)
+	{
+	    XFreeCursor (display->display, s->busyCursor);
+	    s->busyCursor = XCreateFontCursor (display->display, XC_watch);
+
+	    XFreeCursor (display->display, s->normalCursor);
+	    s->normalCursor = XCreateFontCursor (display->display, XC_left_ptr);
+#ifdef HAVE_XINPUT2
+	    XIDefineCursor (display->display, clientPointer,
+			    s->root, s->normalCursor);
+#else
+	    XDefineCursor (display->display, s->root, s->normalCursor);
+#endif
+	    XFlush (display->display);
+	}
+    }
+}
+
 const CompMetadataOptionInfo coreDisplayOptionInfo[COMP_DISPLAY_OPTION_NUM] = {
     { "abi", "int", 0, 0, 0 },
     { "active_plugins", "list", "<type>string</type>", 0, 0 },
@@ -510,7 +552,9 @@ const CompMetadataOptionInfo coreDisplayOptionInfo[COMP_DISPLAY_OPTION_NUM] = {
     { "toggle_window_shaded_key", "key", 0, shade, 0 },
     { "ignore_hints_when_maximized", "bool", 0, 0, 0 },
     { "ping_delay", "int", "<min>1000</min>", 0, 0 },
-    { "edge_delay", "int", "<min>0</min>", 0, 0 }
+    { "edge_delay", "int", "<min>0</min>", 0, 0 },
+    { "cursor_theme", "string", 0, 0, 0 },
+    { "cursor_size", "int", 0, 0, 0 }
 };
 
 CompOption *
@@ -661,6 +705,20 @@ setDisplayOption (CompPlugin	  *plugin,
 	    return TRUE;
 	}
 	break;
+    case COMP_DISPLAY_OPTION_CURSOR_THEME:
+	if (compSetStringOption (o, value))
+	{
+	    compDisplaySetCursorTheme (display);
+	    return TRUE;
+	}
+	break;
+    case COMP_DISPLAY_OPTION_CURSOR_SIZE:
+	if (compSetIntOption (o, value))
+	{
+	    compDisplaySetCursorTheme (display);
+	    return TRUE;
+	}
+	break;
     default:
 	if (compSetDisplayOption (display, o, value))
 	    return TRUE;
@@ -1106,9 +1164,11 @@ updateModifierMappings (CompDisplay *d)
 	    index = 0;
 	    do
 	    {
-		keysym = XKeycodeToKeysym (d->display,
-					   d->modMap->modifiermap[i],
-					   index++);
+		/* convert keycode to keysym */
+		keysym = key[(d->modMap->modifiermap[i] - minKeycode) *
+					 keysymsPerKeycode + index];
+
+		index++;
 	    } while (!keysym && index < keysymsPerKeycode);
 
 	    if (keysym)
@@ -1409,6 +1469,12 @@ eventLoop (void)
 			pointerX = event.xclient.data.l[2] >> 16;
 			pointerY = event.xclient.data.l[2] & 0xffff;
 		    }
+		    else
+		    if (event.xclient.message_type == d->wmMoveResizeAtom)
+		    {
+		    lastPointerX = pointerX = event.xclient.data.l[0];
+		    lastPointerY = pointerY = event.xclient.data.l[1];
+		    }
 		default:
 		    break;
 		}
@@ -1469,6 +1535,8 @@ eventLoop (void)
 	    time = timeToNextRedraw;
 	    if (time)
 		time = doPoll (time);
+	    else
+		doPoll (0);
 
 	    if (time == 0)
 	    {
@@ -1493,6 +1561,8 @@ eventLoop (void)
 			    timeDiff = 0;
 
 			makeScreenCurrent (s);
+			/* make sure X is ready for us to draw */
+			glXWaitX ();
 
 			if (s->slowAnimations)
 			{
@@ -1521,7 +1591,8 @@ eventLoop (void)
 				break;
 			    }
 
-			    if (s->damageMask & COMP_SCREEN_DAMAGE_ALL_MASK)
+			    if ((s->damageMask & COMP_SCREEN_DAMAGE_ALL_MASK) &&
+					s->nOutputDev == 1)
 			    {
 				s->damageMask &= ~COMP_SCREEN_DAMAGE_ALL_MASK;
 				s->damageMask |=
@@ -1799,6 +1870,46 @@ freeDisplay (CompDisplay *d)
     free (d);
 }
 
+static Bool
+aquireSelection (CompDisplay *d,
+		 int         screen,
+		 const char  *name,
+		 Atom        selection,
+		 Window      owner,
+		 Time        timestamp)
+{
+    Display *dpy = d->display;
+    Window  root = XRootWindow (dpy, screen);
+    XEvent  event;
+
+    XSetSelectionOwner (dpy, selection, owner, timestamp);
+
+    if (XGetSelectionOwner (dpy, selection) != owner)
+    {
+	compLogMessage ("core", CompLogLevelError,
+			"Could not acquire %s manager "
+			"selection on screen %d display \"%s\"",
+			name, screen, DisplayString (dpy));
+
+	return FALSE;
+    }
+
+    /* Send client message indicating that we are now the manager */
+    event.xclient.type         = ClientMessage;
+    event.xclient.window       = root;
+    event.xclient.message_type = d->managerAtom;
+    event.xclient.format       = 32;
+    event.xclient.data.l[0]    = timestamp;
+    event.xclient.data.l[1]    = selection;
+    event.xclient.data.l[2]    = 0;
+    event.xclient.data.l[3]    = 0;
+    event.xclient.data.l[4]    = 0;
+
+    XSendEvent (dpy, root, FALSE, StructureNotifyMask, &event);
+
+    return TRUE;
+}
+
 Bool
 addDisplay (const char *name)
 {
@@ -2027,6 +2138,7 @@ addDisplay (const char *name)
 	XInternAtom (dpy, "_NET_CLIENT_LIST_STACKING", 0);
 
     d->frameExtentsAtom = XInternAtom (dpy, "_NET_FRAME_EXTENTS", 0);
+    d->gtkFrameExtentsAtom = XInternAtom (dpy, "_GTK_FRAME_EXTENTS", 0);
     d->frameWindowAtom  = XInternAtom (dpy, "_NET_FRAME_WINDOW", 0);
 
     d->wmStateAtom	  = XInternAtom (dpy, "WM_STATE", 0);
@@ -2276,34 +2388,14 @@ addDisplay (const char *name)
 
 	wmSnTimestamp = event.xproperty.time;
 
-	XSetSelectionOwner (dpy, wmSnAtom, newWmSnOwner, wmSnTimestamp);
-
-	if (XGetSelectionOwner (dpy, wmSnAtom) != newWmSnOwner)
+	if (!aquireSelection (d, i, "window", wmSnAtom, newWmSnOwner,
+			      wmSnTimestamp))
 	{
-	    compLogMessage ("core", CompLogLevelError,
-			    "Could not acquire window manager "
-			    "selection on screen %d display \"%s\"",
-			    i, DisplayString (dpy));
-
 	    XDestroyWindow (dpy, newWmSnOwner);
 
 	    continue;
 	}
 
-	/* Send client message indicating that we are now the WM */
-	event.xclient.type	   = ClientMessage;
-	event.xclient.window       = XRootWindow (dpy, i);
-	event.xclient.message_type = d->managerAtom;
-	event.xclient.format       = 32;
-	event.xclient.data.l[0]    = wmSnTimestamp;
-	event.xclient.data.l[1]    = wmSnAtom;
-	event.xclient.data.l[2]    = 0;
-	event.xclient.data.l[3]    = 0;
-	event.xclient.data.l[4]    = 0;
-
-	XSendEvent (dpy, XRootWindow (dpy, i), FALSE,
-		    StructureNotifyMask, &event);
-
 	/* Wait for old window manager to go away */
 	if (currentWmSnOwner != None)
 	{
@@ -2327,15 +2419,9 @@ addDisplay (const char *name)
 	    continue;
 	}
 
-	XSetSelectionOwner (dpy, cmSnAtom, newCmSnOwner, wmSnTimestamp);
-
-	if (XGetSelectionOwner (dpy, cmSnAtom) != newCmSnOwner)
+	if (!aquireSelection (d, i, "compositing", cmSnAtom,
+			      newCmSnOwner, wmSnTimestamp))
 	{
-	    compLogMessage ("core", CompLogLevelError,
-			    "Could not acquire compositing manager "
-			    "selection on screen %d display \"%s\"",
-			    i, DisplayString (dpy));
-
 	    continue;
 	}
 
