@@ -83,11 +83,6 @@ handleSyncAlarm (CompWindow *w)
 {
     if (w->syncWait)
     {
-	if (w->syncWaitHandle)
-	{
-	    compRemoveTimeout (w->syncWaitHandle);
-	    w->syncWaitHandle = 0;
-	}
 
 	w->syncWait = FALSE;
 
@@ -1544,6 +1539,13 @@ handleEvent (CompDisplay *d,
 	    if (w)
 		updateIconGeometry (w);
 	}
+	else if (event->xproperty.atom == d->gtkFrameExtentsAtom)
+	{
+	    w = findWindowAtDisplay (d, event->xproperty.window);
+	    if (w)
+		updateClientFrame (w);
+	}
+
 	else if (event->xproperty.atom == d->winOpacityAtom)
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
