@@ -40,6 +40,9 @@ char *programName;
 char **programArgv;
 int  programArgc;
 
+char **initialPlugins = NULL;
+int  nInitialPlugins = 0;
+
 char *backgroundImage = NULL;
 
 REGION   emptyRegion;
@@ -63,6 +66,7 @@ Bool strictBinding = TRUE;
 Bool noDetection = FALSE;
 Bool useDesktopHints = FALSE;
 Bool onlyCurrentScreen = FALSE;
+Bool noFBO = FALSE;
 static Bool debugOutput = FALSE;
 
 #ifdef USE_COW
@@ -84,6 +88,7 @@ usage (void)
 	    "[--keep-desktop-hints] "
 	    "[--loose-binding] "
 	    "[--replace]\n       "
+	    "[--no-fbo] "
 	    "[--sm-disable] "
 	    "[--sm-client-id ID] "
 	    "[--only-current-screen]\n      "
@@ -154,6 +159,30 @@ logLevelToString (CompLogLevel level)
     return "Unknown";
 }
 
+void
+launchFallbackWM (void)
+{
+    char *fallback = NULL;
+
+    if (strcmp (getenv ("XDG_CURRENT_DESKTOP"), "MATE") == 0)
+	fallback = "marco";
+    else if (strcmp (getenv ("XDG_CURRENT_DESKTOP"), "GNOME") == 0 || strcmp (getenv ("XDG_CURRENT_DESKTOP"), "GNOME") == 0)
+	fallback = "metacity";
+    else if (access ("/usr/bin/xfwm4", F_OK) == 0)
+	fallback = "xfwm4";
+    else if (access ("/usr/bin/icewm", F_OK) == 0)
+	fallback = "icewm";
+
+    printf ("\nLaunching fallback window manager\n");
+    if (fallback != NULL)
+	execlp (fallback, fallback, "--replace", (char *)NULL);
+    else
+	execlp ("xterm", "xterm", (char *)NULL);
+
+    /* we should never get here but if we do just exit */
+    exit (EXIT_FAILURE);
+}
+
 static void
 signalHandler (int sig)
 {
@@ -163,6 +192,9 @@ signalHandler (int sig)
     case SIGCHLD:
 	waitpid (-1, &status, WNOHANG | WUNTRACED);
 	break;
+    case SIGSEGV:
+	launchFallbackWM ();
+	break;
     case SIGHUP:
 	restartSignal = TRUE;
 	break;
@@ -265,6 +297,7 @@ main (int argc, char **argv)
     signal (SIGCHLD, signalHandler);
     signal (SIGINT, signalHandler);
     signal (SIGTERM, signalHandler);
+    signal (SIGSEGV, signalHandler);
 
     emptyRegion.rects = &emptyRegion.extents;
     emptyRegion.numRects = 0;
@@ -343,6 +376,10 @@ main (int argc, char **argv)
 	{
 	    onlyCurrentScreen = TRUE;
 	}
+	else if (!strcmp (argv[i], "--no-fbo"))
+	{
+		noFBO = TRUE;
+	}
 
 #ifdef USE_COW
 	else if (!strcmp (argv[i], "--use-root-window"))
@@ -385,6 +422,16 @@ main (int argc, char **argv)
 	}
     }
 
+    /* add in default plugins if none are given */
+    if (nPlugin == 0)
+    {
+        plugin[nPlugin++] = "ccp";
+        plugin[nPlugin++] = "move";
+        plugin[nPlugin++] = "resize";
+        plugin[nPlugin++] = "place";
+        plugin[nPlugin++] = "decoration";
+    }
+
     if (refreshRateArg)
     {
 	ctx.refreshRateData = malloc (strlen (refreshRateArg) + 256);
@@ -413,6 +460,17 @@ main (int argc, char **argv)
 
 	    ptr += sprintf (ptr, "</default>");
 	}
+
+	initialPlugins = malloc (nPlugin * sizeof (char *));
+	if (initialPlugins)
+	{
+	    memcpy (initialPlugins, plugin, nPlugin * sizeof (char *));
+	    nInitialPlugins = nPlugin;
+	}
+	else
+	{
+	    nInitialPlugins = 0;
+	}
     }
 
     xmlInitParser ();
@@ -445,7 +503,17 @@ main (int argc, char **argv)
     coreInitialized = TRUE;
 
     if (!disableSm)
+    {
+	if (clientId == NULL)
+	{
+	    char *desktop_autostart_id = getenv ("DESKTOP_AUTOSTART_ID");
+	    if (desktop_autostart_id != NULL)
+		clientId = strdup (desktop_autostart_id);
+	    unsetenv ("DESKTOP_AUTOSTART_ID");
+	}
+
 	initSession (clientId);
+    }
 
     if (!addDisplay (displayName))
 	return 1;
@@ -462,6 +530,9 @@ main (int argc, char **argv)
 
     xmlCleanupParser ();
 
+    if (initialPlugins)
+        free (initialPlugins);
+
     if (restartSignal)
     {
 	execvp (programName, programArgv);
